/*
* simple WS2812 LED driver example 
* for Arduino Uno
*
* driver discussed and made by Tim in
* http://www.mikrocontroller.net/topic/292775
*
* addapted for Arduino Uno from 
*
* chris 30.7.2013
*
*/

#define ws2812_port PORTB // Data port register
#define ws2812_pin 2 // Number of the data out pin

void setup() { 
  uint8_t mask;  
  // initialize the digital pin as an output.
  // Pin10 is the led driver data. This Output has to correspond to 
  // the ws2812_pin definition
  pinMode(10, OUTPUT); // pin number on Arduino Uno Board    
}
#define ARRAYLEN 6
uint8_t ledArray[ARRAYLEN];

void loop() 
{
  uint8_t i;

   ledArray[0]=0; 
   ledArray[1]=0;
   ledArray[2]=0; 
   ws2812_sendarray(ledArray,ARRAYLEN);
   delay(500);
   
   ledArray[0]=255; 
   ledArray[1]=0;
   ledArray[2]=0; 
   ws2812_sendarray(ledArray,ARRAYLEN);        
   delay(500);

   ledArray[3]=0; 
   ledArray[4]=0;
   ledArray[5]=32; 
   ws2812_sendarray(ledArray,ARRAYLEN);
   delay(500);
   
   ledArray[3]=32; 
   ledArray[4]=0;
   ledArray[5]=0; 
   ws2812_sendarray(ledArray,ARRAYLEN);        
   delay(500);
}

/*****************************************************************************************************************

    Led Driver Source from
    https://github.com/cpldcpu/light_ws2812/blob/master/light_ws2812_AVR/light_ws2812.c


     light weight WS2812 lib
    
     Created: 07.04.2013 15:57:49 - v0.1
     21.04.2013 15:57:49 - v0.2 - Added 12 Mhz code, cleanup
     07.05.2013 - v0.4 - size optimization, disable irq
     20.05.2013 - v0.5 - Fixed timing bug from size optimization
     27.05.2013 - v0.6 - Major update: Changed I/O Port access to byte writes
     30.6.2013 - V0.7 branch - bug fix in ws2812_sendarray_mask by chris
     
     instead of bit manipulation. This removes this timing
     discrepancy between standard AVR and reduced core so that
     only one routine is required. This comes at the cost of
     additional register usage.
     28.05.2013 - v0.7 - Optimized timing and size of 8 and 12 Mhz routines.
     All routines are within datasheet specs now, except of
     9.6 Mhz which is marginally off.
     03.06.2013 - v0.8 - 9.6 Mhz implementation now within specifications.
     - brvs->brcs. Loops terminate correctly
     Author: Tim (cpldcpu@gmail.com)

*****************************************************************************************************************/
void ws2812_sendarray_mask(uint8_t *, uint16_t , uint8_t);

void ws2812_sendarray(uint8_t *data,uint16_t datlen)
{
  ws2812_sendarray_mask(data,datlen,_BV(ws2812_pin));
}

/*
This routine writes an array of bytes with RGB values to the Dataout pin
using the fast 800kHz clockless WS2811/2812 protocol.
The description of the protocol in the datasheet is somewhat confusing and
it appears that some timing values have been rounded.
The order of the color-data is GRB 8:8:8. Serial data transmission begins
with the most significant bit in each byte.
The total length of each bit is 1.25µs (20 cycles @ 16Mhz)
* At 0µs the dataline is pulled high.
* To send a zero the dataline is pulled low after 0.375µs (6 cycles).
* To send a one the dataline is pulled low after 0.625µs (10 cycles).
After the entire bitstream has been written, the dataout pin has to remain low
for at least 50µs (reset condition).
Due to the loop overhead there is a slight timing error: The loop will execute
in 21 cycles for the last bit write. This does not cause any issues though,
as only the timing between the rising and the falling edge seems to be critical.
Some quick experiments have shown that the bitstream has to be delayed by
more than 3µs until it cannot be continued (3µs=48 cyles).

*/
void ws2812_sendarray_mask(uint8_t *da2wd3ta,uint16_t datlen,uint8_t maskhi)
{
  uint8_t curbyte,ctr,masklo;
  masklo	=~maskhi&ws2812_port;
  maskhi |=ws2812_port;

  while (datlen--) 
  {
    curbyte=*da2wd3ta++; 
  
    asm volatile(
    " ldi %0,8 \n\t"	// 0
    "loop%=:out %2, %3 \n\t"	// 1
    " lsl %1 \n\t"	// 2
    " dec %0 \n\t"	// 3
    
    " rjmp .+0 \n\t"	// 5
    
    " brcs .+2 \n\t"	// 6l / 7h
    " out %2, %4 \n\t"	// 7l / -
    
    " rjmp .+0 \n\t"	// 9
    
    " nop \n\t"	// 10
    " out %2, %4 \n\t"	// 11
    " breq end%= \n\t"	// 12 nt. 13 taken
    
    " rjmp .+0 \n\t"	// 14
    " rjmp .+0 \n\t"	// 16
    " rjmp .+0 \n\t"	// 18
    " rjmp loop%= \n\t"	// 20
    "end%=: \n\t"
    :	"=&d" (ctr)
    :	"r" (curbyte), "I" (_SFR_IO_ADDR(ws2812_port)), "r" (maskhi), "r" (masklo)
    );
  }
}


